!$Id: viirs_nasa_read_module.f90 1541 2016-02-26 23:18:41Z dbotambekov $
!--------------------------------------------------------------------------------------
! Clouds from AVHRR Extended (CLAVR-x) 1b PROCESSING SOFTWARE Version
! 5.4
!
! NAME: viirs_nasa_read_module.f90
!
! PURPOSE: VIIRS NASA (NetCDF4) read tool
!
! DESCRIPTION:  This module deals with reading VIIRS NASA data
!
! AUTHORS:
!  Andrew Heidinger, Andrew.Heidinger@noaa.gov
!  Yue Li, CIMSS, yue.li@ssec.wisc.edu
!  Steve Wanzong, CIMSS, 
!
! COPYRIGHT
! THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
! DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
! FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
! INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
! EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
! DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
! THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
! SUPPORT TO USERS.
!
!--------------------------------------------------------------------------------------

module FY3D_READ_MODULE 

  use READH5DATASET, only: &
   h5readattribute &
   , h5readdataset
  
  use FILE_UTILITY, only: &
      get_lun  
  use FILE_TOOLS, only: &
        FILE_SEARCH &
      , GETLUN
  use PIXEL_COMMON_MOD, only: &
        Sensor &
      , Image &
      , Gap_Pixel_Mask &
      , Temp_Pix_Array_1
  use CONSTANTS_MOD, only: &
        Real4 &
      , Real8 &
      , Int8 &
      , Int4 &
      , Int2 &
      , Int1 &
      , Sym &
      , Exe_Prompt &
      , Missing_Value_Real4 &
      , Missing_Value_Int4 
   use CLAVRX_MESSAGE_MOD
   use CALIBRATION_CONSTANTS_MOD, only: &
        sat_name &
      , solar_ch20 &
      , solar_ch20_nu &
      , ew_ch20 &
      , planck_a1 &
      , planck_a2 &
      , planck_nu
   USE hdf5

  implicit none

  public :: READ_FY3D_DATA
  public :: READ_FY3D_DATE_TIME
  public :: READ_NUMBER_OF_SCANS_FY3D
  public :: READ_FY3D_INSTR_CONSTANTS
  private :: DETERMINE_FY3D_FILE
  private :: JULIAN
  private :: CONVERT_FY3D_RADIANCE
  private :: CONVERT_RAD_2_SOL_REF_DNB

  character(len=18), parameter:: FY3D_PROMPT="FY3D_MODULE:"
 
  ! - bowtie gaps values
  integer, parameter :: Ny_Pattern = 16
  integer, parameter :: Nx_Pattern = 3200
  integer(kind=int4), dimension(:,:), allocatable, public :: Gap_Line_Idx_Pattern
  integer(kind=int1), dimension(:,:), allocatable, public :: Gap_Pixel_Mask_Pattern
  
  character(len=13), parameter:: MODULE_PROMPT="FY3D_MODULE:"

      contains

!--------------------------------------------------------------------
! read viirs nasa data
!--------------------------------------------------------------------
subroutine READ_FY3D_DATA (Segment_Number, VGEOM_File, Error_Out)

   use Pixel_Common_Mod , only : &
        Geo &
      , Nav &
      , Ancil_Data_Dir &
      , Ch &
      , Bt_11um_Sounder &
      , Bt_12um_Sounder 

   use PLANCK_MOD

   use VIEWING_GEOMETRY_MOD, only: &
        GLINT_ANGLE &
      , SCATTERING_ANGLE &
      , RELATIVE_AZIMUTH

   use CALIBRATION_CONSTANTS_MOD, only: &
        Planck_Nu &
      , Planck_Nu_11um_Sndr &
      , Planck_Nu_12um_Sndr

  ! USE hdf5

      integer(kind=int4), intent(in):: Segment_Number
      character(len=*), intent(in):: VGEOM_File
      integer(kind=int4), intent(out):: Error_Out

      integer(kind=int4) :: Nx_Start , Nx_End , Ny_Start , Ny_Start_Iband , Ny_End , N_Seg_Lines
      integer(kind=int4) :: I_Geo
      integer(kind=int4) :: Mband_Start
      integer(kind=int4) :: I_band
      integer(kind=int4) :: N_Mband
      integer(kind=int4) :: I_Iband
      integer(kind=int4) :: N_Iband
      integer(kind=int4) :: I_Fusion
      integer(kind=int4) :: N_Fusion
      integer(kind=int4) :: Fusion_Start
      integer(kind=int4) :: k , i
      integer(kind=int4) :: Lun
      integer(kind=int4) :: Io_Err_Stat
      integer(kind=int8) :: N_Valid
      integer(kind=int4), dimension(2) :: Dim_Seg
      integer(kind=int4), dimension(3) :: Dim_Seg1
      integer(kind=int4), dimension(2) :: Dim_Seg_Iband
      integer(kind=int4), dimension(2) :: Dim_Dnb_Seg
      integer(kind=int4), dimension(2) :: Offset_Mband
      integer(kind=int4), dimension(3) :: Offset_Mband1
      integer(kind=int4), dimension(2) :: Offset_Iband
      integer(kind=int4), dimension(3200) :: D2m_Idx
      integer(kind=int4), dimension(:), allocatable :: Time_Sec_Day
      integer(kind=int4), dimension(7) :: Scaled_Geo = [0, 0, 0, 1, 1, 1, 1]
      integer(kind=int4), dimension(21) :: Modis_Chn_List
      integer(kind=int4), dimension(5) :: Modis_I_Chn_List
      real(kind=real4) :: Fill_Value
      real(kind=real4) :: Scale_Factor , Add_Offset
      real(kind=real4), dimension(:), pointer :: Scale_Factor1 , Add_Offset1
      !real(kind=real4), dimension(4) :: Scale_Factor1, Add_Offset1
      real(kind=real8), parameter :: Sec_Per_Day = 86400.
      real(kind=real8), dimension(:), pointer :: R1d_Buffer
      real(kind=real4), dimension(:,:), pointer :: R2d_Buffer
      real(kind=real4), dimension(:,:), allocatable :: R2d_Temp
      real(kind=real4), dimension(:,:), allocatable :: Iband_Temp_Buff
      real(kind=real4), dimension(:,:), allocatable :: Iband_Temp_Buff_Bt
      integer(kind=int4), dimension(:,:), pointer :: I2d_Buffer
      integer(kind=int4), dimension(:,:,:), pointer :: I2d_Buffer1
      real(kind=real4), dimension(:,:,:), pointer :: R2d_Buffer1
      character(len=1020) :: File_2_Read
      character(len=1020) :: File_Dnb_Idx
      character(len=2) :: Band_Num_Str
      character(len=2) :: Fusion_Band_Str
      character(len=5) :: Day_Night_Flag
      character(len=100), dimension ( 7 ) :: Setname_Geo_List = (/ character (len =300) :: &
                           'Geolocation/Latitude             ' & ! 1
                         , 'Geolocation/Longitude            ' & ! 2
                         , 'Timedata/Millisecond_Count  ' & ! 3
                       !  , 'scan_line_attributes/scan_start_time  ' & ! 3
                         , 'Geolocation/SensorAzimuth       ' & ! 4
                         , 'Geolocation/SensorZenith        ' & ! 5
                         , 'Geolocation/SolarAzimuth        ' & ! 6
                         , 'Geolocation/SolarZenith         ' & ! 7
                                                    /)
      character(len=100) :: Setname_Name
      logical, dimension(21) :: Is_Mband_On
      logical, dimension(21) :: Is_band_Read
      logical, dimension(5) :: Is_Iband_On
      logical, dimension(5) :: Is_Iband_Read
      logical :: Read_Iband_Flag
      !real(kind=real4), dimension(2,6) :: Rad_Coef
      !real(kind=real4), dimension(2,Num_Chan_Sensor) :: Rad_Coef
      real(kind=real4), dimension(:,:,:), allocatable :: Rad_Coef_IR
      real(kind=real4), dimension(:,:), allocatable :: Rad_Coef_VIS

!!!
      real(kind=4),    dimension( 200, 4,  6) :: var_real4_ir
      !real(kind=4),    dimension( 3,  6)                  :: var_real4_ir
      integer ::   error 
      integer (HSIZE_T), dimension(2)         :: dims_geo2
      integer (HID_T)  :: file_id              ! file id 
      integer (HID_T) :: sds_id_var           ! sds id all variables
      !real(kind=4) :: H5T_NATIVE_REAL

!!!

      Error_Out = 0

      ! --- calculate start and end segment limits
      !Number_Of_Lines_Per_Segment
      !Number_Of_Lines_Read_This_Segment
      !Number_Of_Lines

      Nx_Start = 1
      Nx_End = Nx_Start + Image%Number_Of_Elements - 1

      Ny_Start = (Segment_Number - 1) * Image%Number_Of_Lines_Per_Segment + 1
      Ny_End = min(Image%Number_Of_Lines, Ny_Start + Image%Number_of_Lines_Per_Segment - 1)
      N_Seg_Lines = Ny_End - Ny_Start + 1


      Offset_Mband = [Nx_Start - 1, Ny_Start - 1]
      Dim_Seg = [Nx_End - Nx_Start + 1, Ny_End - Ny_Start + 1]
      
!print *, 'Offset_Mband ',Nx_Start - 1, Ny_Start - 1
!print *, 'Dim_Seg ',Nx_End - Nx_Start + 1, Ny_End - Ny_Start + 1
   
      ! --- read geo file data
      ! loop over geo data
      do I_Geo = 1, 7
        if ( I_Geo == 3 ) then
           call H5READDATASET (trim(Image%Level1b_Path)//trim(VGEOM_File), &
                      trim(Setname_Geo_List(I_Geo)), R2d_Buffer)
        else
           call H5READDATASET (trim(Image%Level1b_Path)//trim(VGEOM_File), &
                      trim(Setname_Geo_List(I_Geo)), Offset_Mband, Dim_Seg, R2d_Buffer)
           if (Scaled_Geo(I_Geo) .eq. 1) then
              call H5READATTRIBUTE (trim(Image%Level1b_Path)//trim(VGEOM_File), &
                      trim(Setname_Geo_List(I_Geo)//'/Slope'), Scale_Factor)
              call H5READATTRIBUTE (trim(Image%Level1b_Path)//trim(VGEOM_File), &
                      trim(Setname_Geo_List(I_Geo)//'/Intercept'), Add_Offset)
              call H5READATTRIBUTE (trim(Image%Level1b_Path)//trim(VGEOM_File), &
                      trim(Setname_Geo_List(I_Geo)//'/FillValue'), Fill_Value)
              where ( R2d_Buffer .ne. Fill_Value)
                 R2d_Buffer = (R2d_Buffer * Scale_Factor) + Add_Offset
              endwhere
           endif
           !print *,'I_Geo ',I_Geo
        endif

        ! save read data to output in correct format
        select case (I_Geo)
         case(1)
            Nav % Lat_1b (:,1:N_Seg_Lines) = R2d_Buffer
         case(2)
            Nav % Lon_1b (:,1:N_Seg_Lines) = R2d_Buffer
         case(3)
!print *,'size R2d ',size(R2d_Buffer),shape(R2d_Buffer),shape(R2d_Buffer(0,:)),shape(R2d_Buffer(:,0))
            allocate (Time_Sec_Day (size(R2d_Buffer(0,:))))
            Time_Sec_Day = int((mod(R2d_Buffer(0,:)/1000, Sec_Per_Day)) * 1000)

            ! make data missing of missing scan time
            where (R2d_Buffer(0,:) < 0)
               Time_Sec_Day = Missing_Value_Int4
            endwhere

            Image%Scan_Time_Ms(1:N_Seg_Lines) = (/(Time_Sec_Day((k - 1) / 16 + 1), &
                                              k = Ny_Start, Ny_End)/)
            deallocate ( Time_Sec_Day )
         case(4)
            Geo % Sataz (:,1:N_Seg_Lines) = R2d_Buffer
         case(5)
            Geo % Satzen (:,1:N_Seg_Lines) = R2d_Buffer
         case(6)
            Geo % Solaz (:,1:N_Seg_Lines) = R2d_Buffer
         case(7)
            Geo % Solzen (:,1:N_Seg_Lines) = R2d_Buffer
         end select

         deallocate ( R2d_Buffer)
         !if (I_Geo == 3) deallocate ( R1d_Buffer)

      enddo

      ! --- calculate ascending/descending
      Nav % Ascend (1:N_Seg_Lines) = 0
      do i = 1, N_Seg_Lines - 1
         if (Nav%Lat_1b (Dim_Seg(1) / 2, i + 1) <= Nav%Lat_1b(Dim_Seg(1) / 2, i)) &
                    Nav % Ascend( i ) = 1
      end do
      ! --- fix for the last line
      if ( Nav%Lat_1b(Nx_End / 2, N_Seg_Lines) <= &
           Nav%Lat_1b(Nx_End / 2, N_Seg_Lines - 1) ) &
                Nav % Ascend (N_Seg_Lines) = 1

      ! --- calculate relative azimuth
      Geo % Relaz = RELATIVE_AZIMUTH (Geo%Solaz, Geo%Sataz)

      !--- compute the glint zenith angle
      Geo % Glintzen = GLINT_ANGLE (Geo%Solzen, Geo%Satzen, Geo%Relaz )

      !--- compute the scattering angle
      Geo % Scatangle = SCATTERING_ANGLE (Geo%Solzen, Geo%Satzen, Geo%Relaz)


      ! --- read m-band data
      ! find m-band file
      if (Sensor%Platform_Name == 'FY3D') then
        call DETERMINE_FY3D_FILE(trim(Image%Level1b_Path),trim(VGEOM_File), &
                                    '1000M',File_2_Read)
      endif


      ! --- if no file quit
      if (trim(File_2_Read) .eq. 'no_file') then
         Error_Out = 1
         return
      endif

      ! --- read global attribute Day_Night_Flag
      !!!!! Note: if Day_Night_Flag = 'Night' NO M01 - M06, M09, and M11 bands  !!!!!
      !!!!! If Day_Night_Flag = 'Day' or 'Both' then ALL M-bands exist         !!!!!
      call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
                                       'Day Or Night Flag', Day_Night_Flag)


      ! --- mapping modis to viirs
      !                 041 044 048 055 068  074  085 124 138  160 225 375  405  855  108  120
      !                 M1  M2  M3  M4  M5   M6   M7  M8  M9   M10 M11 M12  M13  M14  M15  M16
      !Modis_Chn_List = [ 8 , 9 , 3 , 4 , 1 , 15 , 2 , 5 , 26 , 6 , 7 , 20 , 22 , 29 , 31 , 32 ]
      Modis_Chn_List = [3,4,1,2,26,6,7,8,9,10,15,17,18,19,5,20,23,28,29,31,32]
!      Modis_Chn_List = [3,4,1,2,26,6,7,8,9,10,0,0,0,15,0,17,18,19,5,20,23,28,29,31,32]

      Is_Mband_On = Sensor%Chan_On_Flag_Default (Modis_Chn_List) == sym%YES
      N_Mband = 21 

      Setname_Name = "Calibration/IR_Cal_Coeff"
         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
                             Setname_Name, R2d_Buffer1)
      !print *,'ir coef shape ', shape(R2d_Buffer1)

    !  if (.not. allocated(Rad_Coef_IR))  allocate(Rad_Coef_IR(6,4,200))
      if (.not. allocated(Rad_Coef_IR))  allocate(Rad_Coef_IR(200,4,6))
     Rad_Coef_IR = R2d_Buffer1
     deallocate (R2d_Buffer1)
!print *, 'new ir coef ',Rad_Coef_IR(1,:,:)
      Setname_Name = "Calibration/VIS_Cal_Coeff"
         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
                             Setname_Name, R2d_Buffer)
!      print *,'vis coef shape', shape(R2d_Buffer)
!      print *, R2d_buffer
      if (.not. allocated(Rad_Coef_VIS))  allocate(Rad_Coef_VIS(3,19))
     Rad_Coef_VIS = R2d_Buffer
!     print *,'new'
!     print *,  rad_coef_vis
     deallocate (R2d_Buffer)


! Chan 1-4 (FY3D Ch 1-4, MODIS Ch 3,4,1,2)
! Chan 5-15 (FY3D Ch 5-10, 14, 16-19, MODIS Ch 26, 6-10, 15, 17-19, 5)
! Chan 16-19 (FY3D Ch 20-23, MODIS Ch 20, 23, 28,29)
! Chan 20-21 (FY3D Ch 24-25, MODIS Ch31, 32)

      do I_band = 1,21
         if (I_band >=1 .and. I_band <=4) then
            setname_Name = "Data/EV_250_Aggr.1KM_RefSB"
         elseif (I_band >=5 .and. I_band <=15 ) then
            setname_Name = "Data/EV_1KM_RefSB"
         elseif (I_band >=16 .and. I_band <=19) then
            setname_Name = "Data/EV_1KM_Emissive"
         elseif (I_band >=20 .and. I_band <=21) then
            setname_Name = "Data/EV_250_Aggr.1KM_Emissive"
         endif

         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
                              Setname_Name//'/Slope', Scale_Factor1)
         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
                              Setname_Name//'/Intercept', Add_Offset1)
         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
                              Setname_Name//'/FillValue', Fill_Value)
         Offset_Mband1 = [Nx_Start - 1, Ny_Start - 1, 0]
         Dim_Seg1 = [Nx_End - Nx_Start + 1, Ny_End - Ny_Start + 1, shape(Scale_Factor1)]
      !
         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
                             Setname_Name, Offset_Mband1, Dim_Seg1, I2d_Buffer1)

         ! - check if channel is on 
         Is_band_Read(I_band) = .false.
         if (.not. Is_Mband_On(I_band)) cycle
         print *,'I_band ',I_band,Modis_Chn_List(I_band)

         allocate (R2d_Temp(Dim_Seg(1),Dim_Seg(2)))
         R2d_Temp = Missing_Value_Real4

! VIS bands
       if (Day_Night_Flag /= 'N') then
         if (I_band >=1 .and. I_band <=4) then
         !    allocate (R2d_Temp(Dim_Seg(1),Dim_Seg(2)))
         !    R2d_Temp = Missing_Value_Real4
             Ch (Modis_Chn_List(I_band)) % Ref_Toa (:, 1:N_Seg_Lines) = Missing_Value_Real4
             R2d_Temp =  float(I2d_Buffer1(:,:,I_band)) 
             ! - unscale
             where ( I2d_Buffer1(:,:,I_band) .ne. Fill_Value)
                   Ch (Modis_Chn_List(I_band)) % Ref_Toa (:, 1:N_Seg_Lines) = &
!                                 ( (I2d_Buffer1(:,:,I_band) * Scale_Factor1(I_band)) + Add_Offset1(I_band) )*100
                          Rad_Coef_VIS(1,I_band)             +Rad_Coef_VIS(2,I_band)*R2d_Temp+ &
                          Rad_Coef_VIS(3,I_band)*R2d_Temp**2
             endwhere
             !print *, 'rad_temp range',maxval(R2d_Temp),minval(R2d_Temp)
             !print *, 'ref range ',maxval(Ch(Modis_Chn_List(I_band))%Ref_Toa(:,1:N_Seg_Lines)), &
             !   minval(Ch(Modis_Chn_List(I_band))%Ref_Toa(:,1:N_Seg_Lines))
         elseif (I_band >= 5 .and. I_band <=15) then
         !    allocate (R2d_Temp(Dim_Seg(1),Dim_Seg(2)))
         !    R2d_Temp = Missing_Value_Real4
             Ch (Modis_Chn_List(I_band)) % Ref_Toa (:, 1:N_Seg_Lines) = Missing_Value_Real4

             ! - unscale
             if (I_band < 11) then   ! number 1 - 6
                R2d_Temp =  float(I2d_Buffer1(:,:,I_band-4))
                where ( I2d_Buffer1(:,:,I_band-4) .ne. Fill_Value)
             !          R2d_Temp =  float(I2d_Buffer1(:,:,I_band)
                       Ch (Modis_Chn_List(I_band)) % Ref_Toa (:, 1:N_Seg_Lines) = &
!                                    ( (I2d_Buffer1(:,:,I_band-4) * Scale_Factor1(I_band-4)) + Add_Offset1(I_band-4) )*100
                          Rad_Coef_VIS(1,I_band)             + Rad_Coef_VIS(2,I_band)*R2d_Temp+ &
                          Rad_Coef_VIS(3,I_band)*R2d_Temp**2
                endwhere
             endif

             if (I_band == 11) then  ! number 10
                R2d_Temp = float(I2d_Buffer1(:,:,10))
                where ( I2d_Buffer1(:,:,10) .ne. Fill_Value)
                      Ch(Modis_Chn_List(I_band)) % Ref_Toa (:, 1:N_Seg_Lines) = &
!                                    ( (I2d_Buffer1(:,:,10) * Scale_Factor1(10)) + Add_Offset1(10) )*100
                          Rad_Coef_VIS(1,10)             + Rad_Coef_VIS(2,10)*R2d_Temp+ &
                          Rad_Coef_VIS(3,10)*R2d_Temp**2
                endwhere
             endif

             if (I_band > 11) then  ! number 12 - 15
                R2d_Temp =  float(I2d_Buffer1(:,:,I_band))
                where ( I2d_Buffer1(:,:,I_band) .ne. Fill_Value)
                      Ch(Modis_Chn_List(I_band)) % Ref_Toa (:, 1:N_Seg_Lines) = &
!                                    ( (I2d_Buffer1(:,:,I_band) * Scale_Factor1(I_band)) + Add_Offset1(I_band) )*100
                          Rad_Coef_VIS(1,I_band)             + Rad_Coef_VIS(2,I_band)*R2d_Temp+ &
                          Rad_Coef_VIS(3,I_band)*R2d_Temp**2
                endwhere
             endif
         endif
       endif

! IR bands
         if (I_band >=16 .and. I_band <=21) then
         !    allocate (R2d_Temp(Dim_Seg(1),Dim_Seg(2)))
         !    R2d_Temp = Missing_Value_Real4

             if (I_band >=16 .and. I_band <=19) then
                !print *, 'shape b16-19 ',shape(I2d_Buffer1)
                Ch (Modis_Chn_List(I_band)) % Rad_Toa (:, 1:N_Seg_Lines) = Missing_Value_Real4
                where ( I2d_Buffer1(:,:,I_band-15) .ne. Fill_Value)
                      R2d_Temp =  (I2d_Buffer1(:,:,I_band-15)+Add_Offset1(I_band-15)) * Scale_Factor1(I_band-15) 
                      !R2d_Temp = I2d_Buffer1(:,:,I_band-15)
                      Ch(Modis_Chn_List(I_band)) % Rad_Toa (:, 1:N_Seg_Lines) = &
                         R2d_Temp
!                         (R2d_Temp-Rad_Coef_IR(100,1,I_band-15))*Rad_Coef_IR(100,2,I_band-15)
!                         Rad_Coef_IR(100,1,I_band-15)             + Rad_Coef_IR(100,2,I_band-15)*R2d_Temp+ &
!                         Rad_Coef_IR(100,3,I_band-15)*R2d_Temp**2 + Rad_Coef_IR(100,4,I_band-15)*R2d_Temp**3
!                     Ch (Modis_Chn_List(I_band)) % Rad_Toa (:, 1:N_Seg_Lines) = &
!                                   (I2d_Buffer1(:,:,I_band-15)-Add_Offset1(I_band-15)) * Scale_Factor1(I_band-15)
!                                   ( (I2d_Buffer1(:,:,I_band-15) * Scale_Factor1(I_band-15)) + Add_Offset1(I_band-15) )!*1E-2
                endwhere

             elseif (I_band >=20 .and. I_band <=21) then
                !print *, 'shape b20-21 ',shape(I2d_Buffer1)
                Ch (Modis_Chn_List(I_band)) % Rad_Toa (:, 1:N_Seg_Lines) = Missing_Value_Real4
                where ( I2d_Buffer1(:,:,I_band-19) .ne. Fill_Value)
                      R2d_Temp =  (I2d_Buffer1(:,:,I_band-19)+Add_Offset1(I_band-19)) *  Scale_Factor1(I_band-19) 
                      !R2d_Temp = I2d_Buffer1(:,:,I_band-19)
                      Ch(Modis_Chn_List(I_band)) % Rad_Toa (:, 1:N_Seg_Lines) = &
                         R2d_Temp    ! direct seems work (opt = 21)
!                         (R2d_Temp-Rad_Coef_IR(100,1,I_band-15))*Rad_Coef_IR(100,2,I_band-15)
!                         Rad_Coef_IR(100,1,I_band-15)             + Rad_Coef_IR(100,2,I_band-15)*R2d_Temp+ &
!                         Rad_Coef_IR(100,3,I_band-15)*R2d_Temp**2 + Rad_Coef_IR(100,4,I_band-15)*R2d_Temp**3
                endwhere
       
             endif

        !     deallocate (R2d_Temp)

             call CONVERT_FY3D_RADIANCE (Ch(Modis_Chn_List(I_band))%Rad_Toa(:,1:N_Seg_Lines), &
                             Planck_Nu(Modis_Chn_List(I_band)), Missing_Value_Real4)

             ! - calculate bt
             call COMPUTE_BT_ARRAY (Ch(Modis_Chn_List(I_band))%Bt_Toa(:,1:N_Seg_Lines), &
                                    Ch(Modis_Chn_List(I_band))%Rad_Toa(:,1:N_Seg_Lines), &
                                    Modis_Chn_List(I_band), Missing_Value_Real4)

             !print *, 'bt range',maxval(Ch(Modis_Chn_List(I_band))%Bt_Toa(:,1:N_Seg_Lines)), &
             !   minval(Ch(Modis_Chn_List(I_band))%Bt_Toa(:,1:N_Seg_Lines))
             !print *, 'rad range ',maxval(Ch(Modis_Chn_List(I_band))%Rad_Toa(:,1:N_Seg_Lines)), &
             !   minval(Ch(Modis_Chn_List(I_band))%Rad_Toa(:,1:N_Seg_Lines))

         endif

        deallocate (R2d_Temp)
        Is_band_Read(I_band) = .true.
        deallocate (I2d_Buffer1)

      enddo

! open geo hdf5 file
call h5open_f(error)
call h5fopen_f (trim(Image%Level1b_Path)//trim(File_2_Read), H5F_ACC_RDONLY_F, file_id, error)

call h5dopen_f (file_id, '/Calibration/IR_Cal_Coeff', sds_id_var, error)
call h5dread_f (sds_id_var, H5T_NATIVE_REAL, var_real4_ir, dims_geo2, error)
!print *, 'var_real4_ir ',var_real4_ir(1,:,:)

      ! --- global variables which have to be set
      Image%Number_Of_Lines_Read_This_Segment = N_Seg_Lines
      do i = 1, Image%Number_Of_Lines_Per_Segment
         Image%Scan_Number(i) = Ny_Start + i - 1
      end do

end subroutine READ_FY3D_DATA

!--------------------------------------------------------------------
! read viirs nasa time
!--------------------------------------------------------------------
subroutine READ_FY3D_DATE_TIME (Path, Infile, Year , Doy , Start_Time &
                , End_Time , Orbit , End_Year, End_Doy )

      character(len=*), intent(in) :: Path
      character(len=*), intent(in) :: Infile
      integer, intent(out) :: Year
      integer, intent(out) :: Doy    !day of year
      integer, intent(out) :: Start_Time  !millisec
      integer, intent(out) :: End_Time    !millisec
      integer, intent(out) :: Orbit
      integer , intent(out) :: End_Year
      integer, intent(out) :: End_Doy    !day of year

      character(len=35) :: Tmp_String 
      integer :: Month
      integer :: Day
      integer :: Start_Hour
      integer :: Start_Minute
      integer :: Start_Sec

      integer :: End_Hour
      integer :: End_Minute
      integer :: End_Sec


      ! --- read time and date from the attributes
      call H5ReadAttribute( trim(Path)//trim(Infile), &
           'Orbit Number', Orbit)

      call H5ReadAttribute( trim(Path)//trim(Infile), &
           'Observing Beginning Time', Tmp_String)
      !0        1         2
      !123456789012345678901234567890
      !2016-04-20T12:00:00.000Z
      !read(Tmp_String(1:4), fmt="(I4)") Year
      !read(Tmp_String(6:7), fmt="(I2)") Month
      !read(Tmp_String(9:10), fmt="(I2)") Day
      read(Tmp_String(1:2), fmt="(I2)") Start_Hour
      read(Tmp_String(4:5), fmt="(I2)") Start_Minute
      read(Tmp_String(7:8), fmt="(I2)") Start_Sec

     call H5ReadAttribute( trim(Path)//trim(Infile), &
           'Observing Beginning Date', Tmp_String)

      read(Tmp_String(1:4), fmt="(I4)") Year
      read(Tmp_String(6:7), fmt="(I2)") Month
      read(Tmp_String(9:10), fmt="(I2)") Day

      !--- compute start day of year
      call JULIAN ( Day, Month, Year, Doy )
      
      call H5ReadAttribute( trim(Path)//trim(Infile), &
           'Observing Ending Time', Tmp_String)
      !read(Tmp_String(1:4), fmt="(I4)") End_Year
      !read(Tmp_String(6:7), fmt="(I2)") Month
      !read(Tmp_String(9:10), fmt="(I2)") Day
      read(Tmp_String(1:2), fmt="(I2)") End_Hour
      read(Tmp_String(4:5), fmt="(I2)") End_Minute
      read(Tmp_String(7:8), fmt="(I2)") End_Sec

     call H5ReadAttribute( trim(Path)//trim(Infile), &
           'Observing Ending Date', Tmp_String)

      read(Tmp_String(1:4), fmt="(I4)") End_Year
      read(Tmp_String(6:7), fmt="(I2)") Month
      read(Tmp_String(9:10), fmt="(I2)") Day

      !--- compute end day of year
      call JULIAN ( Day, Month, End_Year, End_Doy )

      ! --- Calculate start and end time
      Start_Time = ((Start_Hour * 60 + Start_Minute) * 60 + Start_Sec) * 1000
      End_Time = ((End_Hour * 60 + End_Minute) * 60 + End_Sec) * 1000

end subroutine READ_FY3D_DATE_TIME

!--------------------------------------------------------------------
! read viirs nasa number of scans
!--------------------------------------------------------------------
subroutine READ_NUMBER_OF_SCANS_FY3D (Infile, Number_Of_Viirs_Lines, &
                 Error_Out)

      character(len=*), intent(in) :: Infile
      integer(kind=int4), intent(out) :: Error_Out
      integer(kind=int4), intent(out) :: Number_of_Viirs_Lines

      character(len=100) :: Setname
      integer, dimension(:), pointer :: Test
      integer, dimension(1) :: Dims

      integer :: Number_of_Scans     


      Error_Out = 0
      !Setname = 'scan_line_attributes/scan_quality'

      call H5ReadAttribute( trim(Infile), &
           'Number Of Scans', Number_of_Scans)
!      call H5ReadDataset( trim(Infile), trim(Setname), Test )
      Dims = Number_of_Scans
!print *,'Dims ',Dims
      ! --- error 
      if (Dims(1) .lt. 1) then
         Number_Of_Viirs_Lines = -999
         Error_Out = 1
         Return
      endif
         
      Number_Of_Viirs_Lines = Dims(1) * 10

end subroutine READ_NUMBER_OF_SCANS_FY3D

!--------------------------------------------------------------------
! find viirs nasa files
!--------------------------------------------------------------------
subroutine DETERMINE_FY3D_FILE(Path_In,File_In,File_Type_In,File_Out)

      character(len=*), intent(in):: Path_In
      character(len=*), intent(in):: File_In
      character(len=*), intent(in):: File_Type_In
      character(len=*), intent(out):: File_Out

      character(len=500):: Search_String
      character(len=1020), dimension(:), pointer:: Files
      integer(kind=int4):: Num_Files


      !0        1         2         3         4         5
      !12345678901234567890123456789012345678901234567890
      !VGEOM_snpp_d20160420_t120000_c20160420175142.nc
      !VL1BM_snpp_d20160420_t120000_c20160420172929.nc
      !VGEOD_snpp_d20140205_t115400_c20170405013130.nc
      !VL1BD_snpp_d20140205_t115400_c20170405012902.nc

      if (index(File_Type_In, '1000M') > 0) then 
         Search_String = trim(File_In(1:33))//trim(File_Type_In)//'*.HDF'
      !0        1         2         3         4         5
      !12345678901234567890123456789012345678901234567890
      !VNP03MOD.A2014036.1154.001.2017300063048.uwssec.nc
      !VNP02MOD.A2014036.1154.001.2017300062642.uwssec.nc
      !VNP03DNB.A2014036.1154.001.2017300063048.uwssec.nc
      !VNP02DNB.A2014036.1154.001.2017300062642.uwssec.nc
      endif

      Files => FILE_SEARCH(trim(Path_In),trim(Search_String),count=Num_Files)

      if (Num_Files == 0) then
         !print *, EXE_PROMPT, FY3D_PROMPT, " No "//trim(File_Type_In)//" File Found"
         !call MESG(" No "//trim(File_Type_In)//" File Found", level = verb_lev % DEFAULT)
         File_Out = "no_file"
         return
      endif

      if (Num_Files > 1) then
         !print *, EXE_PROMPT, FY3D_PROMPT, "Multiple "//trim(File_Type_In)//" Files Found"
         !call MESG(" Multiple "//trim(File_Type_In)//" Files Found", level = verb_lev % DEFAULT)
!         File_Out = "no_file"
      endif

      File_Out = Files(1)
      print *, EXE_PROMPT, FY3D_PROMPT, trim(File_Type_In)//" File Found, ",trim(File_Out)

      Files => null()

end subroutine DETERMINE_FY3D_FILE

!-------------------------------------------------
! subroutine JULIAN(iday,imonth,iyear,jday)
! compute julian day
! input:
!         iday - integer day
!         imonth - integer month
!         iyear - integer year (2 or four digits)
!         
! output : jday - julian day
!--------------------------------------------------
subroutine JULIAN(iday,imonth,iyear,jday)

        integer, intent(in)::  iday,imonth,iyear
        integer, intent(out):: jday
        integer::  j
        integer, dimension(12)::  jmonth

        jmonth = reshape ((/31,28,31,30,31,30,31,31,30,31,30,31/),(/12/))

        jday = iday
        if (modulo(iyear,4) == 0) then
            jmonth(2)=29
        endif

        do j = 1,imonth-1
           jday = jday + jmonth(j)
        end do

end subroutine JULIAN

!--------------------------------------------------
! Function Name: CONVERT_VIIRS_RADIANCE
!
! Function:
!    Convert to units of the VIIRS radiance values from the that used
!    in the IDPS level-1b to that expected by CLAVR-x
!
! Description: 
!   
! Calling Sequence: rad_new =
! convert_viirs_radiance(rad_old,nu,missing_value)
!   
!
! Inputs:
!   rad_old = radiance in units of W/m^2/micron/str (2d array)
!   nu = channels equivalent width in units of cm^-1
!   missing_value = value assigned to missing radiance values
!
! Outputs: 
!   rad_new = radiance in units of mW/m^2/cm^-1/str (2d array)
!
! Dependencies:
!
! Restrictions:  None
!
! Reference: algebraic manipulation of Planck Equation
! ---------------------------------------------------------------------------------------
subroutine CONVERT_FY3D_RADIANCE(Radiance,Nu,Missing_Value)
      real (kind=real4), dimension(:,:), intent(inout):: Radiance
      real (kind=real4), intent(in):: Nu
      real (kind=real4), intent(in):: Missing_Value

      where(Radiance /= Missing_Value)
!         Radiance = Radiance * (((10000.0 / Nu )**2) / 10.0)
         Radiance = Radiance !/ 10000.
      end where

      return

end subroutine CONVERT_FY3D_RADIANCE

!--------------------------------------------------
subroutine CONVERT_RAD_2_SOL_REF_DNB( Radiance, Solzen , Day_Of_Year , &
                                      Missing_Value , Reflectance )

      real (kind=real4), dimension(:,:), intent(in) :: Radiance
      real (kind=real4), dimension(:,:), intent(in) :: Solzen
      integer(kind=int2), intent(in) :: Day_Of_Year
      real (kind=real4), intent(in) :: Missing_Value
      real (kind=real4), dimension(:,:), intent(out) :: Reflectance
      real (kind=real4), parameter :: Fo = 0.044217282   ! dnb solar energy in w/cm2 ( Source? )
      real , parameter :: PI = 3.14159265359
      real , parameter :: DTOR = PI / 180.
      real :: Sun_Earth_Distance

      Sun_Earth_Distance = 1.0 - 0.016729 * cos( 0.9856 * ( Day_Of_Year - 4.0 ) * DTOR )

      Reflectance = Missing_Value

      where(Radiance /= Missing_Value)
         Reflectance = 100. * (Pi * Radiance * Sun_Earth_Distance ** 2) / &
                            (cos(Solzen * DTOR) * Fo)
      end where

end subroutine CONVERT_RAD_2_SOL_REF_DNB

!-------------------------------------------------------------------------------------
! subroutine to compute the bowtie gap pattern.  
! This pattern repeats every 48 scans 
! all VIIRS files should be integer multiples of this pattern
!
! Gap_Pixel_Mask_Pattern = a binary mask that identifies these bowtie gaps 
! Gap_Line_Idx = line index for each pixel in pattern including gap pixels
!
! A description of the pattern
!
!--------  line type 3
!----      line type 4
! 12 lines without gaps
!----      line type 1
!--------  line type 2
!--------  line type 3
!----      line type 4
! 12 lines without gaps
!----      line type 1
!--------  line type 2
!--------  line type 3
!----      line type 4
! 12 lines without gaps
!----      line type 1
!--------  line type 2
!
!  line types 1 and 4 have 640 missing pixels
!  line types 2 and 3 have 1008 missing pixels
! 
!-------------------------------------------------------------------------------------
   subroutine COMPUTE_VIIRS_BOWTIE_GAP_PATTERN()

      integer (kind=int4), dimension(Ny_Pattern):: Line_Type
      integer (kind=int4), parameter:: Ngap_1 = 640  !1280
      integer (kind=int4), parameter:: Ngap_2 = 1008 !2016
      integer (kind=int4), parameter:: Ngap_3 = 1008 !2016
      integer (kind=int4), parameter:: Ngap_4 = 640  !1280
      integer (kind=int4):: Iline
      integer (kind=int4):: i1
      integer (kind=int4):: i2

      !--- define the line patterns as described above
      Line_Type = &
      (/3,4,0,0,0,0,0,0,0,0,0,0,0,0,1,2/)


      if (.not. allocated(Gap_Line_Idx_Pattern)) &
           allocate(Gap_Line_Idx_Pattern(Nx_Pattern,Ny_Pattern))
      if (.not. allocated(Gap_Pixel_Mask_Pattern)) &
           allocate(Gap_Pixel_Mask_Pattern(Nx_Pattern,Ny_Pattern))

      do Iline = 1 , Ny_Pattern

         Gap_Line_Idx_Pattern( : , Iline ) = -999
         Gap_Pixel_Mask_Pattern( : , Iline ) = 0

         if (Line_Type(Iline) == 1) then

            i1 = 1
            i2 = Ngap_1
            Gap_Line_Idx_Pattern( i1 : i2 , Iline) = Iline - 1
            Gap_Pixel_Mask_Pattern( i1 : i2 , Iline) = 1

            i1 = Nx_Pattern - Ngap_1 + 1
            i2 = Nx_Pattern
            Gap_Line_Idx_Pattern( i1 : i2 , Iline ) = Iline - 1
            Gap_Pixel_Mask_Pattern( i1 : i2 , Iline ) = 1
         end if

         if (Line_Type(Iline) == 2) then
            i1 = 1
            i2 = Ngap_1
            Gap_Line_Idx_Pattern( i1 : i2 , Iline ) = Iline - 2
            Gap_Pixel_Mask_Pattern( i1 : i2 , Iline ) = 1

            i1 = Ngap_1 + 1
            i2 = Ngap_2
            Gap_Line_Idx_Pattern( i1 : i2 , Iline ) = Iline - 1
            Gap_Pixel_Mask_Pattern( i1 : i2 , Iline ) = 1

            i1 = Nx_Pattern - Ngap_1 + 1
            i2 = Nx_Pattern
            Gap_Line_Idx_Pattern( i1 : i2 , Iline ) = Iline - 2
            Gap_Pixel_Mask_Pattern( i1 : i2 , Iline ) = 1

            i1 = Nx_Pattern - Ngap_2 + 1
            i2 = i1  + (Ngap_2 - Ngap_1)
            Gap_Line_Idx_Pattern( i1 : i2 , Iline ) = Iline - 1
            Gap_Pixel_Mask_Pattern( i1 : i2 , Iline ) = 1
         end if

         if (Line_Type(Iline) == 3) then
            i1 = 1
            i2 = Ngap_4
            Gap_Line_Idx_Pattern(i1:i2,Iline) = Iline + 2
            Gap_Pixel_Mask_Pattern(i1:i2,Iline) = 1

            i1 = Ngap_4 + 1
            i2 = Ngap_3
            Gap_Line_Idx_Pattern(i1:i2,Iline) = Iline + 1
            Gap_Pixel_Mask_Pattern(i1:i2,Iline) = 1

            i1 = Nx_Pattern - Ngap_4 + 1
            i2 = Nx_Pattern
            Gap_Line_Idx_Pattern(i1:i2,Iline) = Iline + 2
            Gap_Pixel_Mask_Pattern(i1:i2,Iline) = 1

            i1 = Nx_Pattern - Ngap_3 + 1
            i2 = i1 + (Ngap_3 - Ngap_4)
            Gap_Line_Idx_Pattern(i1:i2,Iline) = Iline + 1
            Gap_Pixel_Mask_Pattern(i1:i2,Iline) = 1
         end if

         if (Line_Type(Iline) ==  4) then
            i1 = 1
            i2 = Ngap_4
            Gap_Line_Idx_Pattern(i1:i2,Iline) = Iline + 1
            Gap_Pixel_Mask_Pattern(i1:i2,Iline) = 1

            i1 = Nx_Pattern - Ngap_4 + 1
            i2 = Nx_Pattern
            Gap_Line_Idx_Pattern( i1 : i2 , Iline ) = Iline + 1
            Gap_Pixel_Mask_Pattern(i1:i2,Iline) = 1
         end if

      end do

   end subroutine COMPUTE_VIIRS_BOWTIE_GAP_PATTERN

!------------------------------------------------------------------------------
! this routine uses the bowtie gap pattern, applies it to an arbitrary 
! segment of data and fills in the observations with the closest valid data
!------------------------------------------------------------------------------
    subroutine FILL_VIIRS_BOWTIE_GAPS ( Line_Start, Number_of_Lines  , Variable )
      integer(kind=int4), intent(in) :: Line_Start
      integer(kind=int4), intent(in) :: Number_of_Lines
      real(kind=real4), dimension(:,:), intent (inout) :: Variable

      integer(kind=int4) :: Line_Offset
      integer(kind=int4) :: Line_In_Pattern
      integer(kind=int4) :: Line_in_Segment
      integer(kind=int4) :: Line_Idx
      integer(kind=int4) :: Elem_Idx
      integer(kind=int4), dimension(:,:), allocatable :: Gap_Line_Idx

      allocate (Gap_Line_Idx(Nx_Pattern, Number_Of_Lines))
      Gap_Line_Idx = Missing_Value_Int4

      do Line_In_Segment = 1,  Number_Of_Lines

         Line_In_Pattern = mod (Line_Start - 1 + Line_In_Segment , Ny_Pattern)
         if (Line_In_Pattern == 0) Line_In_Pattern = Ny_Pattern

         Line_Offset = Line_In_Segment - Line_In_Pattern

         if (Gap_Line_Idx_Pattern(1,Line_In_Pattern) /= Missing_Value_Int4) &
             Gap_Line_Idx(:, Line_In_Segment) = Gap_Line_Idx_Pattern(:,Line_In_Pattern) + Line_Offset

         where ( Gap_Line_Idx(:,Line_in_Segment) <= 0 )
            Gap_Line_Idx(:,Line_in_Segment) = 1
         end where

         where ( Gap_Line_Idx(:,Line_in_Segment) > Number_of_Lines )
            Gap_Line_Idx(:,Line_in_Segment) = Number_of_Lines
         end where

         ! - write gap mask to output
         Gap_Pixel_Mask(:,Line_in_Segment) = Gap_Pixel_Mask_Pattern(:,Line_in_Pattern)

         ! - fill gaps
         do Elem_Idx = 1, Nx_Pattern
            if (Gap_Pixel_Mask(Elem_Idx,Line_in_Segment) == 1) then
               Line_Idx = Gap_Line_Idx(Elem_Idx,Line_in_Segment)

               Variable(Elem_Idx, Line_In_Segment) = Variable(Elem_Idx,Line_Idx)

            end if
         end do
      end do

      deallocate (Gap_Line_Idx)

   end subroutine FILL_VIIRS_BOWTIE_GAPS


!----------------------------------------------------------------
! read the fy3d constants into memory
!-----------------------------------------------------------------
   subroutine READ_FY3D_INSTR_CONSTANTS(Instr_Const_File)
     character(len=*), intent(in):: Instr_Const_File
     integer:: ios0, erstat
     integer:: Instr_Const_lun

     Instr_Const_lun = GET_LUN()

     open(unit=Instr_Const_lun,file=trim(Instr_Const_File),status="old",position="rewind",action="read",iostat=ios0)

     print *, EXE_PROMPT, MODULE_PROMPT, " Opening ", trim(Instr_Const_File)
     erstat = 0
     if (ios0 /= 0) then
        erstat = 19
        print *, EXE_PROMPT, MODULE_PROMPT, "Error opening FY3D constants file, ios0 = ", ios0
        stop 19
     endif

     read(unit=Instr_Const_lun,fmt="(a4)") sat_name
     read(unit=Instr_Const_lun,fmt=*) Solar_Ch20
     read(unit=Instr_Const_lun,fmt=*) Ew_Ch20
     read(unit=Instr_Const_lun,fmt=*) planck_a1(20), planck_a2(20), planck_nu(20)
     read(unit=Instr_Const_lun,fmt=*) planck_a1(23), planck_a2(23), planck_nu(23)
     read(unit=Instr_Const_lun,fmt=*) planck_a1(28), planck_a2(28), planck_nu(28)
     read(unit=Instr_Const_lun,fmt=*) planck_a1(29), planck_a2(29), planck_nu(29)
     read(unit=Instr_Const_lun,fmt=*) planck_a1(31), planck_a2(31), planck_nu(31)
     read(unit=Instr_Const_lun,fmt=*) planck_a1(32), planck_a2(32), planck_nu(32)
     close(unit=Instr_Const_lun)

     !-- convert solar flux in channel 20 to mean with units mW/m^2/cm^-1
     Solar_Ch20_Nu = 1000.0 * Solar_Ch20 / Ew_Ch20

   end subroutine READ_FY3D_INSTR_CONSTANTS


   !---------------------------------------------------------------------------
   !!!! Taken from "viirs_clavrx_bridge.f90" which was written by Andi W. !!!
   !!!! Only deleted St.Dev.      Denis B.
   !---------------------------------------------------------------------------
   ! - iband has full file dimension of 6400 x1536
   ! - mband 3200 x 768
   !  - output of min_val ... is 3200 768
   !---------------------------------------------------------------------------
   subroutine COMPUTE_IBAND_STATISTICS ( iband_array , out_min_val, out_max_val , out_mean_val )

      implicit none

      real, dimension(:,:) , intent(in) :: iband_array
      real, dimension(:,:) , intent(out)  :: out_min_val, out_max_val , out_mean_val

      real, dimension(2,2) :: small_iband
      integer :: im , jm
      integer , dimension(2) ::  dim_m
      integer :: iband_x0, iband_x1 ,  iband_y0, iband_y1


      dim_m = shape ( out_min_val )

      out_min_val =-999.
      out_max_val =-999.
      out_mean_val =-999.

      do im = 1 , dim_m( 1)

         iband_x0 = ( im-1) * 2 +1
         iband_x1 = iband_x0 + 1
         do jm =1 , dim_m( 2 )
            iband_y0 = ( jm-1) * 2 +1
            iband_y1 = iband_y0 + 1
            small_iband = iband_array ( iband_x0 :  iband_x1 ,  iband_y0 : iband_y1 )
            if ( minval ( small_iband ) > 0 ) then
               out_min_val ( im, jm ) = minval ( small_iband )
               out_max_val ( im, jm ) = maxval ( small_iband )
               out_mean_val ( im, jm ) = sum ( small_iband ) / 4.0
            end if
         end do
      end do

   end subroutine COMPUTE_IBAND_STATISTICS

!subroutine READ_FY3D_LEVEL1B(path,file_name,iband, &
!                              calibrated_data_out, &
!                              uncert_data_out, &
!                              nx,ny,Seg_Idx,ny_total,ny_local_temp, &
!                              Error_Status)

!        Setname_name = "Data/EV_1KM_Emissive"
         ! - read data
!         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
!                             Setname_Name, Offset_Mband, Dim_Seg, I2d_Buffer)

         ! - read attribute to unscale
!         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
!                              Setname_Name//'/Slope', Scale_Factor)
!         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
!                              Setname_Name//'/Intercept', Add_Offset)
!         call H5READDATASET (trim(Image%Level1b_Path)//trim(File_2_Read), &
!                              Setname_Name//'/FillValue', Fill_Value)
!end subroutine READ_FY3D_LEVEL1B

!====================================================================


end module FY3D_READ_MODULE 


